/**
 * DESCRIPTION OF THE CLASS
 *
 * @author Martin Beseda
 * @author Martin Mrovec
 * @author Michal Kravčenko
 * @date 2017 - 2018
 */

#ifndef INC_4NEURO_NEURONLINEAR_H
#define INC_4NEURO_NEURONLINEAR_H

#include "Neuron.h"

/**
 * Linear neuron class - uses activation function in the form f(x)=a*x + b,
 * 'x' being the neuron's potential
 */
class NeuronLinear:Neuron {
public:

    /**
     * Constructs the object of the Linear neuron with activation function
     * f(x) = b * x + a
     * @param[in] a First coefficient, stored in activation_function_parameters[0]
     * @param[in] b Second coefficient, stored in activation_function_parameters[1]
     */
    explicit NeuronLinear(double a = 0.0, double b = 0.0);

    /**
     * Evaluates 'b*x + a' and stores the result into the 'state' property
     */
    void activate( ) override;

    /**
     * Calculates the partial derivative of the activation function
     * f(x) = b*x + a
     * @param[in] param_idx Index of the parameter to calculate derivative of
     * @return Partial derivative of the activation function according to the
     * 'param_idx'-th parameter. For 'param_idx'=0 returns 'x', otherwise
     * returns 1
     */
    double activation_function_get_partial_derivative(int param_idx) override;

    /**
     * Calculates d/dx of (b*x + a)
     * @return b
     */
    double activation_function_get_derivative( ) override;
};


#endif //INC_4NEURO_NEURONLINEAR_H

/*
!-----------------------!----------------------------------------------------------------------
    ! class linear_neuron_t !
    !-----------------------!

    !> Linear neuron class - uses activation function in the form f(x)=a*x + b,
    !! 'x' being the neuron's potential
    type, extends(neuron_t) :: linear_neuron_t
        private

        !! Coefficients for the linear activation function in format 'f(x)=a*x + b'
        real(kind=real_4neuro) :: a_coef  !< The coefficient 'a' in the activation function f(x)=a*x + b
        real(kind=real_4neuro) :: b_coef  !< The coefficient 'b' in the activation function f(x)=a*x + b

        contains

            !> Activation function - f(x)=a*x + b
            procedure, private :: activate => identity_potential_activate_impl
    end type linear_neuron_t

    interface linear_neuron_t
        !> Non-parametric constructor of linear_neuron_t class - a coef. is 1, b coef. is 0
        !! @return An instance of the class linear_neuron_t
        module procedure :: new_linear_neuron

        !> Constructor of linear_neuron_t class
        !! @param[in] a_coef a coef. of the linear activation function
        !! @param[in] b_coef b coef. of the linear activation fucntion
        !! @return An instance of the class linear_neuron_t
        module procedure :: new_linear_neuron_2
    end interface linear_neuron_t
 */