!> Module containing classes representing connections (synapses)
!! in neural networks.
!!
!! It uses 'neuron_m' and 'time_measurement_m' modules.
!!
!! @author Martin Beseda
!! @author Martin Mrovec
!! @date 2017
!! @todo Rewrite pass_signal method of interval_connection_t 
module connection_m
    use neuron_m
    use time_measurement_m
    use normal_m
    use container_m

    implicit none

    public

    !------------------!------------------------------------------------------------------------
    ! Type definitions !
    !------------------!

    !> Represents a connection between two neurons.
    type, extends(abstract_base_t) :: connection_t
        !TODO properties are public because of problems with reading from binary
        !file - solve in a better way 

        class(neuron_t), pointer :: input_neuron   !< Pointer to an input neuron
        class(neuron_t), pointer :: output_neuron  !< Pointer to an output neuron
        real(kind=real_4neuro)   :: weight         !< Weight of the connection

    contains

        !> Initializes the common connection_t class components
        !! 'input_neuron', 'output_neuron' and 'weight'.
        !! I.e. serves similarly to an abstract constructor.
        !! @param[in] input_neuron Pointer to the input neuron (instance of neuron_t)
        !! @param[in] output_neuron Pointer to the output neuron (instance of neuron_t)
        !! @param[in] weight Weight of the connection (real number)
        procedure, private :: init_components    => connection_init_components_impl

        !> Nullifies pointers to input and output neurons, so they
        !! don't get destroyed when the connection_t instance is 
        !! deallocated.
        procedure, private :: nullify_pointers   => nullify_pointers_impl

        !> Adds a given value to the current weight of the  
        !! connection.
        !! @param[in] added_value Number (real) to be added to the current weight
        procedure          :: adjust_weight      => adjust_weight_impl

        !> Getter for the private 'input_neuron' component
        !! @return Pointer to the input neuron (type neuron_t, pointer)
        procedure          :: get_input_neuron   => get_input_neuron_impl

        !> Getter for the private 'output_neuron' component
        !! @return Pointer to the output neuron (type neuron_t, pointer)
        procedure          :: get_output_neuron  => get_output_neuron_impl

        !> Getter for the private 'weight' component
        !! @return Weight of the connection (type real)
        procedure          :: get_weight         => get_weight_impl

        !> Passes (assigns) the product (input neuron state * weight)
        !! to an output neuron.
        procedure          :: pass_signal        => pass_signal_impl

        !> Implementation of write function enabling the storage of
        !! allocatable arrays.
        procedure          :: write_sample       => write_connection_sample

        !> Implementation of read function enabling to read
        !! the whole instance of net_t stored as binary file.
        procedure          :: read_sample        => read_connection_sample

        generic            :: write(unformatted) => write_sample
        generic            :: read(unformatted)  => read_sample

        !> Scalar desctructor for single instances of the class connection_t
        final              :: destroy_connection

        !> Array desctructor for arrays of instances of the class connection_t
        final              :: destroy_connection_array

    end type connection_t

    interface connection_t
        !> Constructor of connection_t class
        !! @param[in] input_neuron Pointer to the input neuron (instance of neuron_t)
        !! @param[in] output_neuron Pointer to the output neuron (instance of neuron_t)
        !! @return Pointer to the instance of the class connection_t
        module procedure :: new_connection_2

        !> Constructor of connection_t class
        !! @param[in] input_neuron Pointer to the input neuron (instance of neuron_t)
        !! @param[in] output_neuron Pointer to the output neuron (instance of neuron_t)
        !! @param[in] weight Weight of the connection (real number)
        !! @return Pointer to the instance of the class connection_t
        module procedure :: new_connection_3
    end interface connection_t

    !> Represents a connection between two neurons.
    !! Able to pass a signal from an input neuron to 
    !! an output one.
    type, extends(connection_t) :: interval_connection_t
    contains

        !> Passes (assigns) the product 
        !! input neuron state * weight)
        !! to an output neuron.
        !! @todo Rewrite implementation to "interval"
        procedure :: pass_signal => pass_signal_interval_impl

    end type interval_connection_t
    
    interface interval_connection_t
        !> Constructor of interval_connection_t class
        !! @param[in] input_neuron Pointer to the input neuron (instance of neuron_t)
        !! @param[in] output_neuron Pointer to the output neuron (instance of neuron_t)
        !! @return Pointer to the instance of the class interval_connection_t
        module procedure :: new_interval_connection_2 
 
        !> Constructor of interval_connection_t class
        !! @param[in] input_neuron Pointer to the input neuron (instance of neuron_t)
        !! @param[in] output_neuron Pointer to the output neuron (instance of neuron_t)
        !! @param[in] weight Weight of the connection (real number)
        !! @return Pointer to the instance of the class interval_connection_t
        module procedure :: new_interval_connection_3
    end interface interval_connection_t

    contains
        !------------------------!------------------------------------------------------------------
        ! Method implementations !
        ! -----------------------!

        !--------------------!
        ! class connection_t !
        !--------------------!

        !--------------!----------------------------------------------------------------------------
        ! Constructors !
        !--------------!
  
        !> Constructor of connection_t class
        !! @param[in] input_neuron Pointer to the input neuron (instance of neuron_t)
        !! @param[in] output_neuron Pointer to the output neuron (instance of neuron_t)
        !! @return An instance of the class connection_t
        function new_connection_2(input_neuron, output_neuron) result(new_obj)
            class(neuron_t), pointer, intent(in) :: input_neuron
            class(neuron_t), pointer, intent(in) :: output_neuron
            real(kind=real_4neuro)               :: weight
            type(connection_t)                   :: new_obj
            integer(kind=4)                      :: values(8)  !< values(8) is used as seed

#ifdef TIME_PROFILING
            real                                 :: start_time
            call time_profiling_start(start_time)
#endif
            ! Generate pseudorandom number from Gaussian distribution
            ! as connection weight
            call date_and_time(values=values)
            weight = r4_normal_01(values(8))

            call new_obj%init_components(input_neuron, output_neuron, weight)
#ifdef TIME_PROFILING
            call time_profiling_stop(start_time, 'new_connection_2')
#endif
        end function new_connection_2

        !> Constructor of connection_t class
        !! @param[in] input_neuron Pointer to the input neuron (instance of neuron_t)
        !! @param[in] output_neuron Pointer to the output neuron (instance of neuron_t)
        !! @param[in] weight Weight of the connection (real number)
        !! @return An instance of the class connection_t
        function new_connection_3(input_neuron, output_neuron, weight) result(new_obj)
            class(neuron_t), pointer, intent(in) :: input_neuron
            class(neuron_t), pointer, intent(in) :: output_neuron
            real(kind=real_4neuro), intent(in)   :: weight
            type(connection_t)                   :: new_obj

#ifdef TIME_PROFILING
            real                                 :: start_time
            call time_profiling_start(start_time)
#endif
            call new_obj%init_components(input_neuron, output_neuron, weight)
#ifdef TIME_PROFILING
            call time_profiling_stop(start_time, 'new_connection_3')
#endif
        end function new_connection_3

        !-------------!--------------------------------------------------------------------------
        ! Destructors !
        !-------------!
         
        !> Scalar desctructor for single instances of the class connection_t
        subroutine destroy_connection(this)
            type(connection_t), intent(inout) :: this
#ifdef TIME_PROFILING
            real                              :: start_time
            call time_profiling_start(start_time)
#endif
            call this%nullify_pointers()
#ifdef TIME_PROFILING
            call time_profiling_start(start_time, 'destroy_connection')
#endif
        end subroutine destroy_connection

        !> Array desctructor for arrays of instances of the class connection_t
        subroutine destroy_connection_array(this)
            type(connection_t), intent(inout) :: this(:)
            integer(kind=integer_4neuro)      :: i
#ifdef TIME_PROFILING
            real                              :: start_time
            call time_profiling_start(start_time)
#endif
            do i = 1, size(this)
                call this(i)%nullify_pointers()
            end do
#ifdef TIME_PROFILING
            call time_profiling_stop(start_time, 'destroy_connection_array')
#endif
        end subroutine destroy_connection_array

        !-------------------!-------------------------------------------------------------------
        ! Getters & Setters !
        !-------------------!
        
        !> Getter for the private 'input_neuron' component
        !! @return Pointer to the input neuron (type class(*), pointer)
        function get_input_neuron_impl(this) result (input_neuron)
            class(connection_t), target, intent(in) :: this
            class(*), pointer                       :: input_neuron
#ifdef TIME_PROFILING
            real                                    :: start_time
            call time_profiling_start(start_time)
#endif
            input_neuron => this%input_neuron
#ifdef TIME_PROFILING
            call time_profiling_stop(start_time,'get_input_neuron_impl')
#endif
        end function get_input_neuron_impl

        !> Getter for the private 'output_neuron' component
        !! @return Pointer to the output neuron (type class(*), pointer)
        function get_output_neuron_impl(this) result (output_neuron)
            class(connection_t), target, intent(in) :: this
            class(*), pointer                       :: output_neuron
#ifdef TIME_PROFILING
            real                                    :: start_time
            call time_profiling_start(start_time)
#endif
            output_neuron => this%output_neuron
#ifdef TIME_PROFILING
            call time_profiling_stop(start_time, 'get_output_neuron_impl')
#endif
        end function get_output_neuron_impl

        !> Getter for the private 'weight' component
        !! @return Weight of the connection (type real)
        function get_weight_impl(this) result (weight)
            class(connection_t), intent(in) :: this
            real(kind=real_4neuro)          :: weight
#ifdef TIME_PROFILING
            real                            :: start_time
            call time_profiling_start(start_time)
#endif
            weight = this%weight
#ifdef TIME_PROFILING
            call time_profiling_stop(start_time, 'get_weight_impl')
#endif
        end function get_weight_impl


        !----------------!-------------------------------------------------------------------------
        ! Common methods !
        !----------------!

        !> Initializes the common connection_t class components
        !! 'input_neuron', 'output_neuron' and 'weight'.
        !! I.e. serves similarly to an abstract constructor.
        !! @param[in] input_neuron Pointer to the input neuron (instance of neuron_t)
        !! @param[in] output_neuron Pointer to the output neuron (instance of neuron_t)
        !! @param[in] weight Weight of the connection (real number)
        subroutine connection_init_components_impl(this, input_neuron, output_neuron, weight) 
            class(connection_t), intent(inout)   :: this
            class(neuron_t), pointer, intent(in) :: input_neuron
            class(neuron_t), pointer, intent(in) :: output_neuron
            real(kind=real_4neuro), intent(in)   :: weight

#ifdef TIME_PROFILING
            real                               :: start_time
            call time_profiling_start(start_time)
#endif

            this%input_neuron  => input_neuron
            this%output_neuron => output_neuron
            this%weight        = weight

#ifdef TIME_PROFILING
            call connection_time_profiling_stop(start_time, 'connection_init_components_impl')
#endif
        end subroutine connection_init_components_impl

        !> Nullifies pointers to input and output neurons, so they
        !! don't get destroyed when the connection_t instance is 
        !! deallocated.           
        subroutine nullify_pointers_impl(this)
            class(connection_t), intent(inout) :: this
#ifdef TIME_PROFILING
            real                               :: start_time
            call time_profiling_start(start_time)
#endif
            nullify(this%input_neuron)
            nullify(this%output_neuron)
#ifdef TIME_PROFILING
            call time_profiling_stop(start_time, 'nullify_pointers_impl')
#endif
        end subroutine nullify_pointers_impl

        !> Adds a given value to the current weight of the  
        !! connection.
        !! @param[in] added_value Number (real) to be added to the current weight
        subroutine adjust_weight_impl(this, added_value)
            class(connection_t), intent(inout) :: this
            real(kind=real_4neuro), intent(in) :: added_value
#ifdef TIME_PROFILING
            real                               :: start_time
            call time_profiling_start(start_time)
#endif
            this%weight = this%weight + added_value 
#ifdef TIME_PROFILING
            call time_profiling_stop(start_time, 'adjust_weight_impl')
#endif
        end subroutine adjust_weight_impl

        !> Passes (assigns) the product 
        !! input neuron state * weight)
        !! to an output neuron.
        subroutine pass_signal_impl(this)
            class(connection_t), intent(in) :: this
#ifdef TIME_PROFILING
            real                                     :: start_time
            call time_profiling_start(start_time)
#endif
            call this%output_neuron%adjust_potential(this%input_neuron%get_state() * this%weight)
#ifdef TIME_PROFILING
            call time_profiling_stop(start_time, 'pass_signal_impl')
#endif
        end subroutine pass_signal_impl

        !> Implementation of write function enabling the storage of
        !! allocatable arrays.
        !! @param[in] unit Unit (handler) where the file is written
        !! @param[out] iostat Diagnostic value "returned" by the subroutine
        !! @param[inout] iomsg Explaining note about error
        subroutine write_connection_sample(this, unit, iostat, iomsg)
            class(connection_t), target, intent(in) :: this
            integer, intent(in)             :: unit
            integer, intent(out)            :: iostat
            character(*), intent(inout)     :: iomsg
            integer                         :: i      !< Counter
#ifdef TIME_PROFILING
            real                            :: start_time
            call time_profiling_start(start_time)
#endif
            ! Write a record giving sizes for the allocation
            !write(unit, iostat=iostat, iomsg=iomsg) SIZE(this%input_neuron), &
            !                                        SIZE(this%output_neuron)

            write(unit, iostat=iostat, iomsg=iomsg) container_t(this%input_neuron), &
                                                    container_t(this%output_neuron), &
                                                    container_t(this%weight)
#ifdef TIME_PROFILING
            call time_profiling_stop(start_time, 'write_connection_sample')
#endif
        end subroutine write_connection_sample

        ! Unformatted reading for the sample derived type
        subroutine read_connection_sample(this, unit, iostat, iomsg)
            class(connection_t), intent(inout) :: this
            integer, intent(in)                :: unit
            integer, intent(out)               :: iostat
            character(*), intent(inout)        :: iomsg
            integer                            :: i
            class(container_t), allocatable    :: inp_n, out_n
            !integer                            :: s1, s2, s3, s4  !< Sizes of stored allocatable properties
#ifdef TIME_PROFILING
            real                               :: start_time
            call time_profiling_start(start_time)
#endif
            ! We first have a record telling us the sizes of components
            !read(unit, iostat=iostat, iomsg=iomsg) s1, s2, s3, s4

            ! So we do the allocation
            !allocate(this%neuron_arr(s1), &
            !        this%input_neuron_arr(s2), &
            !        this%output_neuron_arr(s3), &
            !        this%connection_arr(s4))

            ! And then finally the reading.
            read(unit, iostat=iostat, iomsg=iomsg) inp_n, &
                                                   out_n, &
                                                   this%weight

#ifdef TIME_PROFILING
            call time_profiling_stop(start_time, 'read_connection_sample')
#endif
    end subroutine read_connection_sample

        !-----------------------------!
        ! class interval_connection_t !
        !-----------------------------!

        !--------------!------------------------------------------------------------------------
        ! Constructors !
        !--------------!

        !> Constructor of interval_connection_t class
        !! @param[in] input_neuron Pointer to the input neuron (instance of neuron_t)
        !! @param[in] output_neuron Pointer to the output neuron (instance of neuron_t)
        !! @return An instance of the class interval_connection_t
        function new_interval_connection_2(input_neuron, output_neuron) result(new_obj)
            class(neuron_t), pointer, intent(in) :: input_neuron
            class(neuron_t), pointer, intent(in) :: output_neuron
            real(kind=real_4neuro)               :: weight
            type(interval_connection_t)          :: new_obj
            integer(kind=4)                      :: values(8)
#ifdef TIME_PROFILING
            real                                 :: start_time
            call time_profiling_start(start_time)
#endif
            call date_and_time(values=values)
            weight = r4_normal_01(values(8))

            call new_obj%init_components(input_neuron, output_neuron, weight)
#ifdef TIME_PROFILING
            call time_profiling_stop(start_time, 'new_interval_connection_2')
#endif
        end function new_interval_connection_2

        !> Constructor of interval_connection_t class
        !! @param[in] input_neuron Pointer to the input neuron (instance of neuron_t)
        !! @param[in] output_neuron Pointer to the output neuron (instance of neuron_t)
        !! @param[in] weight Weight of the connection (real number) 
        !! @return An instance of the class interval_connection_t
        function new_interval_connection_3(input_neuron, output_neuron, weight) result(new_obj)
            class(neuron_t), pointer, intent(in) :: input_neuron
            class(neuron_t), pointer, intent(in) :: output_neuron
            real(kind=real_4neuro), intent(in)   :: weight
            type(interval_connection_t)          :: new_obj
#ifdef TIME_PROFILING
            real                                 :: start_time
            call time_profiling_start(start_time)
#endif
            call new_obj%init_components(input_neuron, output_neuron, weight)

#ifdef TIME_PROFILING
            call time_profiling_stop(start_time, 'new_interval_connection_3')
#endif
        end function new_interval_connection_3

        !----------------!---------------------------------------------------------------------
        ! Common methods !
        !----------------!

        !> Passes (assigns) the product 
        !! input neuron state * weight)
        !! to an output neuron.
        !! @todo Rewrite implementation to "interval"
        subroutine pass_signal_interval_impl(this)
            class(interval_connection_t), intent(in) :: this
#ifdef TIME_PROFILING
            real                                     :: start_time
            call time_profiling_start(start_time)
#endif
            !todo dopsat metodu pro interval
            call this%output_neuron%adjust_potential(this%input_neuron%get_state() * this%weight)
#ifdef TIME_PROFILING
            call time_profiling_stop(start_time, 'pass_signal_interval_impl')
#endif
        end subroutine pass_signal_interval_impl

end module connection_m

