!> Module containing declaration of a container type 'container_t'.
!! This class "wraps" pointers, so it's possible to store them
!! into arrays.
!!
!! @author Martin Beseda
!! @date 2018
module container_m
    use abstract_base_m
    use data_kinds_4neuro_m

    implicit none

    public

    !> Represents a container for a single container_t pointer
    type :: container_t
        ! Variable content is public to make usage of this class with one 
        ! specific purpose as simple as possible.

        class(*), pointer, public :: content

    contains
        !> Scalar desctructor for single instances of the class container_t
        final :: destroy_container

        !> Array desctructor for arrays of instances of the class container_t
        final :: destroy_container_array

    end type container_t

    interface container_t
        !> Constructor of container_t class
        !! @return Instance of the class container_t with nullified content
        module procedure :: new_container_empty

        !> Constructor of container_t class
        !! @param[in] content_in container to be contained (pointer)
        !! @return Pointer to the instance of the class container_t with assigned content
        module procedure :: new_container_assigned
    end interface container_t

contains
        !--------------!------------------------------------------------------------------------
        ! Constructors !
        !--------------!

        !> Constructor of container_t class
        !! @return Instance of the class container_t with nullified content
        function new_container_empty() result(new_obj)
            type(container_t) :: new_obj
#ifdef TIME_PROFILING
            real              :: start_time
            call time_profiling_start(start_time)
#endif
            new_obj%content => null()
#ifdef TIME_PROFILING
            call time_profiling_stop(start_time, 'new_container_empty')
#endif
        end function new_container_empty

        !> Constructor of container_t class
        !! @param[in] content_in connection to be contained (pointer)
        !! @return Pointer to the instance of the class connection_t with assigned content
        function new_container_assigned(content_in) result(new_obj)
            class(abstract_base_t), pointer, intent(in)  :: content_in
            type(container_t)                            :: new_obj
#ifdef TIME_PROFILING
            real                                         :: start_time
            call time_profiling_start(start_time)
#endif
            new_obj%content => content_in
#ifdef TIME_PROFILING
            call time_profiling_stop(start_time, 'new_container_assigned')
#endif
        end function new_container_assigned

        !--------------!------------------------------------------------------------------------
        ! Destructors  !
        !--------------!
        !> Scalar desctructor for single instances of the class container_t
        subroutine destroy_container(this)
            type(container_t), intent(inout) :: this
#ifdef TIME_PROFILING
            real                             :: start_time
            call time_profiling_start(start_time)
#endif
            nullify(this%content)
#ifdef TIME_PROFILING
            call time_profiling_stop(start_time, 'destroy_container')
#endif
        end subroutine destroy_container

        !> Array desctructor for arrays of instances of the class container_t
        subroutine destroy_container_array(this)
            type(container_t), intent(inout)  :: this(:)
            integer(kind=integer_4neuro)      :: i
#ifdef TIME_PROFILING
            real                              :: start_time
            call time_profiling_start(start_time)
#endif
            do i = 1, size(this)
                nullify(this(i)%content)
            end do
#ifdef TIME_PROFILING
            call time_profiling_stop(start_time, 'destroy_container_array')
#endif
        end subroutine destroy_container_array

end module container_m

